"""

"""
import functools
import dataclasses
from collections import OrderedDict
from typing import Optional, Dict
from typing_extensions import Self

from .fields import *
from ..base import Context

#** Variables **#
__all__ = ['sequence', 'make_sequence', 'Sequence']

#: codec-fields field on class objects
FIELDS = '__encoded__'

#: annotations field on class objects
ANNOTATIONS = '__annotations__'

#: placeholder type instead of None to allow for None as a default
MISSING = type('MISSING', (), {})

#** Functions **#

def sequence(cls: Optional[type] = None, **kwargs):
    """
    generate sequence object w/ codec-field and build dataclass-like object

    :param cls:    class object being converted into a sequence
    :param kwargs: kwargs to pass to dataclass generation
    :return:       sequence class object type
    """
    if cls is None:
        return functools.partial(make_sequence, **kwargs)
    return make_sequence(cls, **kwargs)

def make_sequence(cls: type, **kwargs):
    """
    generate a sequence object w/ codec fields and build dataclass-like object

    :param cls:    class object being converted into a sequence
    :param kwargs: kwargs to pass to dataclass generation
    :return:       sequence class object type
    """
    values      = {}
    fields      = getattr(cls, FIELDS, OrderedDict())
    annotations = getattr(cls, ANNOTATIONS, {})
    for name in list(annotations.keys()):
        # skip processing if annotation is a InitVar or ClassVar
        anno = annotations[name]
        if is_datavar(anno):
            continue
        # retrieve default value
        value = getattr(cls, name, MISSING)
        if hasattr(cls, name):
            delattr(cls, name)
        # parse attribute into field if not already
        anno  = compile_annotation(name, anno)
        field = Field(name, anno)
        fields[name] = field
        # process property types
        if isinstance(anno, Property):
            # ensure annotation is deleted for dataclass
            del annotations[name]
            field.type     = anno.hint
            field.dataattr = False
            # validate property function value
            if not isinstance(value, property):
                if not callable(value):
                    raise ValueError(f'property: {name} must be a function')
                value = property(value)
            values[name] = value
            continue
        # process standard codec-types
        if isinstance(anno, Codec):
            annotations[name] = anno.base_type
            values[name]      = value
    # generate bases for new sequence dataclass
    bases = list(cls.__mro__)
    if Sequence not in bases:
        bases.insert(1, Sequence)
    # generate new unique object w/ fields parsed from original
    dataclassfunc = dataclasses.dataclass(**kwargs)
    return dataclassfunc(type(cname(cls), tuple(bases), {
        FIELDS:      fields,
        ANNOTATIONS: annotations,
        **values
    }))

#** Classes **#

class Sequence:
    __encoded__: Dict[str, Field]
  
    def encode(self, ctx: Context) -> bytes:
        """encode the compiled sequence fields into bytes"""
        encoded = bytearray()
        for name, field in self.__encoded__.items():
            # retrieve value for the given attribute
            value = getattr(self, name, MISSING)
            if value is MISSING:
                raise ValueError(f'{cname(self)} missing attr {name!r}')
            # encode it according it's associated codec
            encoded += field.type.encode(ctx, value)
        return bytes(encoded)

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> Self:
        """decode the given raw-bytes into a compiled sequence"""
        kwargs = {}
        for name, field in cls.__encoded__.items():
            kwargs[name] = field.type.decode(ctx, raw)
        return cls(**kwargs)



"""
DNS Question Implementation
"""
from .codec import *
from .enum import RType, RClass

#** Variables **#
__all__ = ['Question', 'Zone']

#** Classes **#

@make_sequence
class Question:
    name:   Domain
    qtype:  Int[16, RType, 'QType']
    qclass: Int[16, RClass, 'QClass'] = RClass.IN

class Zone(Question):
    """Alias of Question in UPDATE action DNS Requests"""
    pass



from enum import IntEnum
from logging import Logger, getLogger
from typing import Optional

from pyserve import Address, Writer
from pyserve import Session as BaseSession

from .backend import Backend
from ..enum import QR, OpCode, RType, RCode
from ..message import Message
from ..edns import EdnsAnswer
from ..exceptions import DnsError, NotImplemented

#** Variables **#
__all__ = ['Session']

#DONE: error on session generation failure on startup of pyserve service
#DONE: ignore server-shutdown failure on udp-sockets (move override to only tcp impl)
#DONE: implement recursive client lookup w/ backend extension
#DONE: simplify record assignment for in-memory backend
#DONE: update server to support inverse-query lookups (PTR)
#DONE: block any messages other than query/inverse-query
#DONE: unify answer types under similar protocol?
#DONE: implement error packet response on server-processing error
#DONE: implement propper logging for dns server impl
#TODO: allow for loading from config for in-memory backend

#** Classes **#

class Mode(IntEnum):
    """allowed runtime modes to run dns server"""
    SYNC           = 1
    ASYNC          = 2
    THREADED       = 3
    THREADED_ASYNC = 4

class Session(BaseSession):
    """Extendable Implementation of DNS Server Session Manager"""
    
    def __init__(self, backend: Backend, logger: Optional[Logger] = None):
        self.logger  = logger or getLogger('pydns')
        self.backend = backend
    
    ### DNS Handlers

    def process_query(self, msg: Message):
        """
        process questions in query message and append answers found
        """
        # process questions include in message
        for q in msg.questions:
            # check if server is authority on domain and retrieve basic answers
            is_authority = self.backend.is_authority(q.name)
            msg.flags.authorative = msg.flags.authorative or is_authority
            answers, source = self.backend.get_answers(q.name, q.qtype)
            # include SOA if authorative and not already included
            if is_authority and q.qtype != RType.SOA and \
                not any(a.name == q.name for a in msg.authority):
                more, _ = self.backend.get_answers(q.name, RType.SOA)
                answers  = answers.copy()
                answers += more
            # report and assign answers
            self.logger.info(f'{self.addr_str} | {q.name} {q.qtype.name} answers={len(answers)} src={source}')
            for answer in answers:
                if answer.rtype == RType.SOA:
                    msg.authority.append(answer)
                else:
                    msg.answers.append(answer)
    
    def process_status(self, msg: Message):
        """
        process dns status request
        """
        raise NotImplemented

    def process_notify(self, msg: Message):
        """
        process dns notify request
        """
        raise NotImplemented

    def process_update(self, msg: Message):
        """
        process dns update request
        """
        raise NotImplemented

    ### Standard Handlers
    
    def connection_made(self, addr: Address, writer: Writer):
        """
        handle session initialization on connection-made
        """
        self.addr     = addr
        self.writer   = writer
        self.addr_str = '%s:%d' % self.addr
        self.logger.debug(f'{self.addr_str} | connection-made')

    def data_recieved(self, data: bytes):
        """
        parse raw packet-data and process request
        """
        self.logger.debug(f'{self.addr_str} | recieved {len(data)} bytes')
        msg = Message.decode(data)
        # ignore request if not a request
        if msg.flags.qr != QR.Question:
            return
        # update flags for response
        msg.flags.qr = QR.Response
        msg.flags.recursion_available = self.backend.recursion_available
        try:
            # add standard empty EDNS response when present
            if any(a.rtype == RType.OPT for a in msg.additional):
                msg.additional = []
                msg.additional.append(EdnsAnswer())
            # reject request if not a query
            if msg.flags.op in (OpCode.Query, OpCode.InverseQuery):
                self.process_query(msg)
            elif msg.flags.op == OpCode.Status:
                self.process_status(msg)
            elif msg.flags.op == OpCode.Notify:
                self.process_notify(msg)
            elif msg.flags.op == OpCode.Update:
                self.process_update(msg)
            else:
                raise NotImplemented
            # process question lookups
        except DnsError as e:
            msg.flags.rcode = e.rcode
        except Exception as e:
            msg.flags.rcode = RCode.ServerFailure
            raise e
        finally:
            # send response
            data = msg.encode()
            self.logger.debug(f'{self.addr_str} | sent {len(data)} bytes')
            self.writer.write(data)

    def connection_lost(self, err: Optional[Exception]):
        """
        debug log connection lost
        """
        self.logger.debug(f'{self.addr_str} | connection-lost err={err}')



import time
import math
import threading
from abc import abstractmethod
from dataclasses import dataclass
from typing import NamedTuple, Protocol, List, Dict

from ...enum import RType
from ...answer import Answer
from ...client import BaseClient
from ...question import Question

#** Variables **#
__all__ = ['Backend', 'Forwarder', 'Cache', 'MemoryBackend']

#** Classes **#

class Answers(NamedTuple):
    """custom return type when retrieving backend answers"""
    answers: List[Answer]
    source:  str

class Backend(Protocol):
    """
    BaseClass Interface Definition for Backend Implementations
    """
    recursion_available: bool = False

    @abstractmethod
    def is_authority(self, domain: bytes) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        raise NotImplementedError

@dataclass(repr=False)
class Forwarder(Backend):
    """
    Recursive Dns-Client Lookup Forwarder when Backend returns no Results
    """
    __slots__ = ('backend', 'client', 'recursion_available')

    backend: Backend
    client:  BaseClient
    
    def __post_init__(self):
        self.recursion_available = True

    def is_authority(self, domain: bytes) -> bool:
        return self.backend.is_authority(domain)

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        query for answers w/ client if base-backend returns empty result
        """
        answers, source = self.backend.get_answers(domain, rtype)
        if not answers:
            source  = self.__class__.__name__
            message = self.client.query(Question(domain, rtype))
            answers.extend(message.answers)
            answers.extend(message.authority)
            answers.extend([
                a for a in message.additional if isinstance(a, Answer)])
        return Answers(answers, source)

@dataclass
class CacheRecord:
    """
    Record Entry for In-Memory Cache
    """
    __slots__ = ('answers', 'expiration', 'ttl', 'lifetime_mod')

    answers:    List[Answer]
    expiration: float
    ttl:        int

    def __post_init__(self):
        self.lifetime_mod = 0

    def lifetime(self):
        """calculate lifetime remaining of record"""
        return int(math.floor(self.expiration - time.time()))

    def ttl_mod(self, lifetime: int) -> int:
        """calculate ttl modifier based on the given lifetime"""
        elapsed = self.ttl - lifetime
        ttl_mod = elapsed - self.lifetime_mod
        self.lifetime_mod = elapsed
        return ttl_mod

@dataclass(repr=False)
class Cache(Backend):
    """
    In-Memory Cache Extension for Backend Results
    """
    backend:    Backend
    expiration: int = 30
    maxsize:    int = 10000

    def __post_init__(self):
        self.recursion_available = self.backend.recursion_available
        self.mutex = threading.Lock()
        self.authorities: Dict[bytes, bool]      = {}
        self.cache:       Dict[str, CacheRecord] = {}
        self.source:      str                    = self.__class__.__name__

    def is_authority(self, domain: bytes) -> bool:
        """
        retrieve if domain is authority from cache before checking backend
        """
        # check cache before querying backend
        if domain in self.authorities:
            return self.authorities[domain]
        # query backend and then permanently cache authority result
        authority = self.backend.is_authority(domain)
        with self.mutex:
            if len(self.authorities) >= self.maxsize:
                self.authorities.clear()
            self.authorities[domain] = authority
        return authority

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        retrieve answers from cache before checking supplied backend
        """
        # attempt to retrieve from cache if it exists
        key     = f'{domain}->{rtype.name}'
        answers = None
        if key in self.cache:
            with self.mutex:
                record   = self.cache[key]
                lifetime = record.lifetime()
                if lifetime > 0:
                    # modify answer TTLs as cache begins to expire
                    expire  = False
                    ttl_mod = record.ttl_mod(lifetime)
                    for answer in record.answers:
                        answer.ttl -= ttl_mod
                        if answer.ttl <= 0:
                            expire = True
                    # only return answers if none are expired
                    if not expire:
                        return Answers(record.answers, self.source)
                del self.cache[key]
        # complete standard lookup for answers
        answers, source = self.backend.get_answers(domain, rtype)
        ttl        = max(a.ttl for a in answers) if answers else self.expiration
        ttl        = min(ttl, self.expiration)
        expiration = time.time() + ttl
        # cache result w/ optional expiration and return results
        with self.mutex:
            if len(self.cache) >= self.maxsize:
                self.cache.clear()
            self.cache[key] = CacheRecord(answers, expiration, ttl)
        return Answers(answers, source)

#** Imports **#
from .memory import MemoryBackend



"""
In-Memory Backend Implementation
"""
import ipaddress
from typing import Dict, List, Set, Any

from . import Backend, Answers, Answer, RType
from ...answer import get_rclass
from ...content import PTR

#** Variables **#
__all__ = ['MemoryBackend']

#: type definiton for in-memory record database
RecordDB = Dict[bytes, Dict[str, List[Answer]]]

#: raw record entry configuration
RecordEntries = Dict[str, List[Dict[str, Any]]]

#** Classes **#

class MemoryBackend(Backend):
    """Simple In-Memory Backend for DNS Records"""
    
    __slots__ = ('records', 'authorities')

    def __init__(self):
        self.records:     RecordDB   = {}
        self.authorities: Set[bytes] = set()
        self.source:      str        = self.__class__.__name__

    def add_answer(self, domain: bytes, answer: Answer):
        """
        add additional domain answer into in-memory db
        """
        rtype = answer.rtype
        self.records.setdefault(domain, {})
        self.records[domain].setdefault(rtype.name, [])
        self.records[domain][rtype.name].append(answer)
 
    def save_domain(self, domain: bytes, entries: RecordEntries):
        """
        save additional records into in-memory db
        """
        # convert dictionary records into valid dns answer objects
        self.authorities.add(domain)
        for rname, records in entries.items():
            rtype  = RType[rname]
            rclass = get_rclass(rtype)
            for record in records:
                # add standard record
                ttl     = record.pop('ttl', 60)
                content = rclass(**record)
                answer  = Answer(domain, ttl, content)
                self.add_answer(domain, answer)
                # reverse ip records for PTR lookups
                ipaddr = record.get('ip', None)
                if ipaddr is not None:
                    ip   = ipaddress.ip_address(ipaddr)
                    name = ip.reverse_pointer.encode()
                    answer = Answer(name, ttl, PTR(domain))
                    self.add_answer(name, answer)

    def is_authority(self, domain: bytes) -> bool:
        """
        retrieve if domain is an authority
        """
        return domain in self.authorities

    def get_answers(self, domain: bytes, rtype: RType) -> Answers:
        """
        retrieve records associated w/ the given domain and record-type
        """
        answers = []
        if domain in self.records:
            records = self.records[domain]
            answers = records.get(rtype.name, [])
        return Answers(answers, self.source)



"""
DNS Client Implementation
"""
from abc import abstractmethod
from random import randint
from typing import Protocol

from ..enum import QR, OpCode
from ..flags import Flags
from ..question import Question
from ..message import Message

#** Variables **#
__all__ = ['BaseClient', 'UdpClient', 'TcpClient', 'HttpsClient']

#** Functions **#

def new_message_id() -> int:
    """
    generate a new valid id for a dns message packet

    :return: new valid message-id integer
    """
    return randint(1, 2 ^ 16)

#** Classes **#

class BaseClient(Protocol):

    @abstractmethod
    def request(self, msg: Message) -> Message:
        raise NotImplementedError

    def query(self, query: Question) -> Message:
        """
        build request message from query and return response

        :param query: simple dns query
        :return:      response message to query
        """
        mid     = new_message_id()
        flags   = Flags(qr=QR.Question, op=OpCode.Query)
        message = Message(id=mid, flags=flags, questions=[query])
        return self.request(message)

#** Imports **#
from .standard import UdpClient, TcpClient
from .https import HttpsClient



"""
Web HTTP Based DNS Client
"""
from urllib.request import Request, urlopen
from dataclasses import dataclass
from typing import Optional

from . import BaseClient, Message

#** Variables **#
__all__ = ['HttpsClient']

#** Classes **#

@dataclass
class HttpsClient(BaseClient):
    url:     str           = 'https://cloudflare-dns.com/dns-query'
    timeout: Optional[int] = None

    def __post_init__(self):
        self.headers = {
            'Accept':       'application/dns-message',
            'Content-Type': 'application/dns-message'
        }

    def request(self, msg: Message) -> Message:
        """
        handle https specific dns request

        :param msg: request message
        :return:    response message
        """
        data    = msg.encode()
        req     = Request(self.url, data=data, headers=self.headers)
        res     = urlopen(req, timeout=self.timeout)
        content = res.read()
        if res.status != 200:
            raise RuntimeError(f'Invalid Response: {res.status} {content}')
        return Message.decode(content)




"""
EDNS OPT Answer Varient Implementation
"""
from dataclasses import dataclass
from typing_extensions import Self

from ..codec import *
from ..enum import RType
from ..answer import BaseAnswer

#** Variables **#
__all__ = ['ROOT', 'EdnsAnswer']

#: root domain (according to EDNS)
ROOT = b''

#** Classes **#

@make_sequence
class Header:
    name:           Domain
    rtype:          Int[16, RType, 'RType']
    payload_size:   Int16
    extended_rcode: Int8
    version:        Int8
    z:              Int16
    data_length:    Int16

@dataclass
class EdnsAnswer(BaseAnswer):
    name:     bytes  = ROOT
    version:  int    = 0
    content:  bytes  = b''
    udp_size: int    = 512

    @property
    def rtype(self) -> RType:
        return RType.OPT

    def encode(self, ctx: Context) -> bytes:
        return Header(
            name=self.name, 
            rtype=self.rtype, 
            payload_size=self.udp_size, 
            extended_rcode=0, 
            version=self.version, 
            z=0, 
            data_length=len(self.content)
        ).encode(ctx)

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> Self:
        header  = Header.decode(ctx, raw)
        content = ctx.slice(raw, header.data_length)
        return cls(
            name=header.name,
            version=header.version,
            udp_size=header.payload_size,
            content=content,
        )




from enum import IntEnum

#** Variables **#
__all__ = [
    'QR',
    'OpCode',
    'RCode',
    'RType',
    'RClass',
    'EDNSOption',
]

#** Classes **#

class QR(IntEnum):
    Question = 0
    Response = 1

class OpCode(IntEnum):
    Query        = 0
    InverseQuery = 1
    Status       = 2
    Notify       = 4
    Update       = 5

class RCode(IntEnum):
    NoError           = 0
    FormatError       = 1
    ServerFailure     = 2
    NonExistantDomain = 3
    NotImplemented    = 4
    Refused           = 5
    YXDomain          = 6
    YXRRSet           = 7
    NXRRSet           = 8
    NotAuthorized     = 9
    NotInZone         = 10

    BadOPTVersion     = 16
    BadSignature      = 16
    BadKey            = 17
    BadTime           = 18
    BadMode           = 19
    BadName           = 20
    BadAlgorithm      = 21

class RType(IntEnum):
    A     = 1
    NS    = 2
    MD    = 3
    MF    = 4
    CNAME = 5
    SOA   = 6
    MB    = 7
    MG    = 8
    MR    = 9
    NULL  = 10
    WKS   = 11
    PTR   = 12
    HINFO = 13
    MINFO = 14
    MX    = 15
    TXT   = 16
    AAAA  = 28
    SRV   = 33
    NAPTR = 35
    OPT   = 41

    DS     = 43
    RRSIG  = 46
    NSEC   = 47
    DNSKEY = 48

    TSIG  = 250

    AXFR  = 252
    MAILB = 253
    MAILA = 254
    ANY   = 255

class RClass(IntEnum):
    IN   = 1
    CS   = 2
    CH   = 3
    HS   = 4
    NONE = 254
    ANY  = 255

class EDNSOption(IntEnum):
    Cookie = 10



"""
Universal Codec Types
"""
from abc import abstractmethod
from dataclasses import dataclass, field
from ipaddress import IPv4Address, IPv6Address
from typing import Dict, Protocol, ClassVar, Callable, Any, Union, Type
from typing_extensions import runtime_checkable

#** Variable **#
__all__ = [
    'Context',

    'Codec',
    'Int',
    'SizedBytes',
    'IpAddress',
    'Domain',
]

#** Classes **#

@dataclass
class Context:
    """Encoding/Decoding Context Tracking"""
    index: int = 0
    index_to_domain: Dict[int, bytes] = field(default_factory=dict)
    domain_to_index: Dict[bytes, int] = field(default_factory=dict)

    def slice(self, raw: bytes, length: int) -> bytes:
        end  = self.index + length
        data = raw[self.index:end]
        self.index = end
        return data

    def save_domain(self, domain: bytes, index: int):
        self.index_to_domain[index] = domain
        self.domain_to_index[domain] = index

@runtime_checkable
class Codec(Protocol):
    """Encoding/Decoding Codec Protocol"""
    base_type: type

    @classmethod
    @abstractmethod
    def encode(cls, ctx: Context, value: Any) -> bytes:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def decode(cls, ctx: Context, raw: bytes) -> Any:
        raise NotImplementedError

class Int(Codec):
    size: ClassVar[int]
    wrap: ClassVar[Callable[[int], Any]]
    base_type: type = int
    
    def __class_getitem__(cls, s: Union[int, tuple]) -> Type['Int']:
        """generate custom Int subclass with the given options"""
        size = s if isinstance(s, int) else s[0]
        wrap = s[1] if isinstance(s, tuple) and len(s) > 1 else lambda x: x
        name = s[2] if isinstance(s, tuple) and len(s) > 2 else cls.__name__
        assert size % 8 == 0, 'size must be multiple of eight'
        cname = f'{name}[{size}]'
        return type(cname, (cls, ), {'size': size // 8, 'wrap': wrap})

    @classmethod
    def encode(cls, ctx: Context, i: int) -> bytes:
        ctx.index += cls.size
        return i.to_bytes(cls.size, 'big')

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> int:
        data = ctx.slice(raw, cls.size)
        size = cls.size * 8
        assert len(data) == cls.size, f'len(bytes)[{len(data)}] != Int[{size}]'
        value = int.from_bytes(data, 'big')
        return cls.wrap(value)

class IpAddress(Codec):
    size:      int
    base_type: Union[Type[IPv4Address], Type[IPv6Address]]

    def __class_getitem__(cls, iptype: str) -> Type['IpAddress']:
        assert iptype in ('ipv4', 'ipv6'), 'invalid ipaddress type'
        size = 4 if iptype == 'ipv4' else 16
        addr = IPv4Address if iptype == 'ipv4' else IPv6Address
        return type(f'IPv{iptype[-1]}', (cls, ), {
            'size': size, 
            'base_type': addr
        })

    @classmethod
    def encode(cls, ctx: Context, ip: Union[str, bytes]) -> bytes:
        packed = cls.base_type(ip).packed
        ctx.index += len(packed)
        return packed

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> Union[IPv4Address, IPv6Address]:
        data = ctx.slice(raw, cls.size)
        return cls.base_type(data)

class SizedBytes(Codec):
    hint:      Type[Codec] = Int[32]
    base_type: type        = bytes
    
    def __class_getitem__(cls, hint: int) -> Type['SizedBytes']:
        codec = Int[hint]
        return type('SizedBytes[{hint!r}]', (cls, ), {'hint': codec})

    @classmethod
    def encode(cls, ctx: Context, content: bytes) -> bytes:
        hint = cls.hint.encode(ctx, len(content))
        ctx.index += len(content)
        return hint + content

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> bytes:
        hint = cls.hint.decode(ctx, raw)
        return ctx.slice(raw, hint)

class Domain(Codec):
    ptr_mask:  int  = 0xC0
    base_type: type = bytes

    @classmethod
    def encode(cls, ctx: Context, domain: bytes) -> bytes:
        encoded = bytearray()
        while domain:
            # check if ptr is an option for remaining domain
            if domain in ctx.domain_to_index:
                index      = ctx.domain_to_index[domain]
                pointer    = index.to_bytes(2, 'big')
                encoded   += bytes((pointer[0] | cls.ptr_mask, pointer[1]))
                ctx.index += 2 
                return bytes(encoded)
            # save partial domain as index
            ctx.save_domain(domain, ctx.index)
            # handle components of name
            split        = domain.split(b'.', 1)
            name, domain = split if len(split) == 2 else (split[0], b'')
            encoded     += len(name).to_bytes(1, 'big') + name
            ctx.index   += 1 + len(name)
        # write final zeros before returning final encoded data
        encoded   += b'\x00'
        ctx.index += 1
        return bytes(encoded)

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> bytes:
        domain = []
        while True:
            # check for length of domain component
            length     = raw[ctx.index]
            ctx.index += 1
            if length == 0:
                break
            # check if name is a pointer
            if length & cls.ptr_mask == cls.ptr_mask:
                name  = bytes((length ^ cls.ptr_mask, raw[ctx.index]))
                index = int.from_bytes(name, 'big')
                base  = ctx.index_to_domain[index]
                domain.append((base, None))
                ctx.index += 1
                break
            # slice name from bytes and updated counter
            idx  = ctx.index - 1
            name = ctx.slice(raw, length)
            domain.append((name, idx))
        # save domain components
        for n, (name, index) in enumerate(domain, 0):
            if index is None:
                continue
            subname = b'.'.join(name for name, _ in domain[n:])
            ctx.save_domain(subname, index)
        return b'.'.join(name for name, _ in domain)



"""
DNS Flags Implementation
"""
from enum import IntFlag
from dataclasses import dataclass
from typing_extensions import Self

from .enum import QR, OpCode, RCode

#** Variables **#
__all__ = ['Flags']

#** Functions **#

def unmask(flags: int, start: int, end: int) -> int:
    """
    unmask a range of bits from the given integer
    """
    m1   = (1 << end) - 1
    m2   = (1 << start) - 1
    mask = m1 ^ m2
    return flags & mask

#** Classes **#

class Flag(IntFlag):
    Authorative      = 1 << 10
    Truncated        = 1 << 9
    RDesired         = 1 << 8
    RAvailable       = 1 << 7
    Authenticated    = 1 << 5
    CheckingDisabled = 1 << 4

@dataclass
class Flags:
    qr:                   QR
    op:                   OpCode
    authorative:          bool  = False
    truncated:            bool  = False
    recursion_desired:    bool  = True
    recursion_available:  bool  = False
    answer_authenticated: bool  = False
    checking_disabled:    bool  = False
    rcode:                RCode = RCode.NoError

    def __int__(self) -> int:
        flags  = self.qr << 15
        flags |= self.op << 11
        flags |= Flag.Authorative if self.authorative else 0
        flags |= Flag.Truncated if self.truncated else 0
        flags |= Flag.RDesired if self.recursion_desired else 0
        flags |= Flag.RAvailable if self.recursion_available else 0
        flags |= Flag.Authenticated if self.answer_authenticated else 0
        flags |= Flag.CheckingDisabled if self.checking_disabled else 0
        flags |= self.rcode
        return flags

    @classmethod
    def fromint(cls, i: int) -> Self:
        return cls(
            qr=QR(i >> 15),
            op=OpCode(unmask(i, 11, 15)),
            authorative=bool(i & Flag.Authorative),
            truncated=bool(i & Flag.Truncated),
            recursion_desired=bool(i & Flag.RDesired),
            recursion_available=bool(i & Flag.RAvailable),
            answer_authenticated=bool(i & Flag.Authenticated),
            checking_disabled=bool(i & Flag.CheckingDisabled),
            rcode=RCode(unmask(i, 0, 4)),
        )



"""
Python Simple DNS Library
"""

#** Variables **#
__all__ = [
    'QR',
    'RType',
    'RClass',

    'Flags',
    'Answer',
    'PreRequisite',
    'Update',
    'Message', 
    'Question',
    'Zone',

    'EdnsAnswer',

    'CNAME',
    'MX',
    'NS',
    'PTR',
    'SOA',
    'A',
    'AAAA',
    'SRV'
]

#** Imports **#

from .flags import Flags
from .enum import QR, RType, RClass
from .answer import Answer, PreRequisite, Update
from .message import Message
from .question import Question, Zone
from .edns import EdnsAnswer
from .content import *



"""
DNS RCode Exceptions
"""
from typing import Dict, Type, Optional, Any

from .enum import RCode

#** Variables **#
__all__ = [
    'make_error',

    'DnsError',
    'ServerFailure',
    'NonExistantDomain',
    'NotImplemented',
]

_EXCEPTIONS: Dict[RCode, Type[Exception]] = {}

#** Functions **#

def make_error(rcode: RCode, message: Any = None):
    """
    retrieve best exception object to match the given rcode
    """
    # cache map of exceptions in module
    global _EXCEPTIONS
    if not _EXCEPTIONS:
        for value in globals().values():
            if isinstance(value, type) and issubclass(value, DnsError):
                _EXCEPTIONS[value.rcode] = value
    # retrieve best-matching exception class based on rcode
    eclass = _EXCEPTIONS.get(rcode, DnsError)
    raise eclass(message, rcode)

#** Classes **#

class DnsError(Exception):
    rcode: RCode = RCode.NoError

    def __init__(self, msg: Any = None, rcode: Optional[RCode] = None):
        self.message = msg
        self.rcode   = rcode or self.rcode

    def __str__(self) -> str:
        if self.message and self.__class__.rcode == self.rcode:
            return str(self.message)
        return super().__str__()

class ServerFailure(DnsError):
    rcode = RCode.ServerFailure

class NonExistantDomain(DnsError):
    rcode = RCode.NonExistantDomain

class NotImplemented(DnsError):
    rcode = RCode.NotImplemented



"""
DNS Encoding/Decoding Helpers
"""
from dataclasses import field, InitVar
from typing import ClassVar

from .base import Context, Codec, Int, SizedBytes, IpAddress, Domain

#** Variables **#
__all__ = [
    'Context',
    'Codec',
    
    'Int',
    'Int8',
    'Int16',
    'Int32',
    'Int48',
    'Int64',

    'IpAddress',
    'Ipv4',
    'Ipv6',

    'Domain',
    'SizedBytes',

    'field',
    'sequence',
    'make_sequence',
    'Sequence',
    'Property',
    'InitVar',
    'ClassVar',
]

Int8  = Int[8]
Int16 = Int[16]
Int32 = Int[32]
Int48 = Int[48]
Int64 = Int[64]

Ipv4 = IpAddress['ipv4']
Ipv6 = IpAddress['ipv6']

#** Imports **#
from .sequence import sequence, make_sequence, Property, Sequence



"""
Sequence Implementation
"""

#** Variables **#
__all__ = [
    'Property',
    'InitVar',
    'ClassVar',
    
    'sequence',
    'make_sequence',
    'Sequence',
]

#** Imports **#
from .fields import Property, InitVar, ClassVar
from .sequence import sequence, make_sequence, Sequence



"""
DNS Message Object
"""
from typing import List
from dataclasses import dataclass, field
from typing import Type
from typing_extensions import Self

from .codec import *
from .enum import OpCode, RType, RCode
from .flags import Flags
from .answer import BaseAnswer, Answer, PreRequisite, Update
from .question import Question, Zone
from .edns import EdnsAnswer
from .exceptions import make_error

#** Variables **#
__all__ = ['Message']

#** Functions **#

def decode_list(cls, num: int, ctx: Context, raw: bytes):
    """
    deserialize a numbered list of items of the specified type
    """
    objects = []
    for _ in range(0, num):
        new = cls.decode(ctx, raw)
        objects.append(new)
    return objects

def decode_answers(cls: Type[Answer], num: int, ctx: Context, raw: bytes):
    """
    deserialize a numbered list of variable answer object types
    """
    objects = []
    for _ in range(0, num):
        # peek the header domain/rtype to determine Answer class
        idx    = ctx.index 
        peek   = PeekHeader.decode(ctx, raw)
        newcls = EdnsAnswer if peek.rtype == RType.OPT else cls
        ctx.index = idx 
        # decode answer class accordingly
        new = newcls.decode(ctx, raw)
        objects.append(new)
    return objects

#** Classes **#

@make_sequence
class PeekHeader:
    name:  Domain
    rtype: Int[16, RType, 'RType']

@make_sequence
class Header:
    id:             Int16
    flags:          Int16 
    num_questions:  Int16
    num_answers:    Int16
    num_authority:  Int16
    num_additional: Int16

@dataclass
class Message:
    id:         int
    flags:      Flags
    questions:  List[Question]   = field(default_factory=list)
    answers:    List[Answer]     = field(default_factory=list)
    authority:  List[Answer]     = field(default_factory=list)
    additional: List[BaseAnswer] = field(default_factory=list)
    
    def raise_on_error(self):
        """raise exception if message contains an error"""
        if self.flags.rcode != RCode.NoError:
            domains = list({q.name for q in self.questions})
            domains = domains[0] if len(domains) == 1 else domains
            make_error(self.flags.rcode, domains or None)

    def encode(self) -> bytes:
        """
        encode the message object into serialized bytes

        :return: serialized message bytes
        """
        ctx = Context()
        raw = bytearray()
        raw += Header(
            id=self.id, 
            flags=int(self.flags), 
            num_questions=len(self.questions), 
            num_answers=len(self.answers), 
            num_authority=len(self.authority), 
            num_additional=len(self.additional)
        ).encode(ctx)
        raw += b''.join(q.encode(ctx) for q in self.questions)
        raw += b''.join(a.encode(ctx) for a in self.answers)
        raw += b''.join(a.encode(ctx) for a in self.authority)
        raw += b''.join(a.encode(ctx) for a in self.additional)
        return bytes(raw)

    @classmethod
    def decode(cls, raw: bytes) -> Self:
        """
        decode the given raw bytes as a DNS message

        :param raw: raw bytes to decode as a DNS message
        :return:    dns message object
        """
        # parse header for message information
        ctx = Context()
        raw = bytes(raw)
        header = Header.decode(ctx, raw)
        flags  = Flags.fromint(header.flags) 
        # determine classes to parse content
        qclass, anclass, auclass = (Question, Answer, Answer) \
                if flags.op != OpCode.Update else \
                (Zone, PreRequisite, Update)
        # parse body content w/ determined classes
        questions  = decode_list(qclass, header.num_questions, ctx, raw)
        answers    = decode_answers(anclass, header.num_answers, ctx, raw)
        authority  = decode_answers(auclass, header.num_authority, ctx, raw)
        additional = decode_answers(Answer, header.num_additional, ctx, raw)
        return cls(
            id=header.id, 
            flags=flags, 
            questions=questions, 
            answers=answers, 
            authority=authority,
            additional=additional,
        )



"""
DNS Standard Content Sequences
"""
from typing import Optional

from .codec import *
from .enum import RType

#** Variables **#
__all__ = [
    'NULL',
    'ANY',
    'CNAME',
    'MX',
    'NS',
    'PTR',
    'SOA',
    'A',
    'AAAA',
    'SRV'
]

#** Functions **#

def content(cls: Optional[type] = None, rtype: Optional[RType] = None):
    """generate content class w/ given rtype"""
    def wrapper(cls):
        cls.rtype = rtype or RType[cls.__name__] 
        return make_sequence(cls)
    return wrapper if cls is None else wrapper(cls)

#** Classes **#

class Content(Sequence):
    rtype: ClassVar[RType]

@content
class NULL(Content):
    pass

@content
class ANY(Content):
    pass

@content
class CNAME(Content):
    name: Domain

@content
class MX(Content):
    preference: Int16
    exchange:   Domain

@content
class NS(Content):
    nameserver: Domain

@content
class PTR(Content):
    ptrname: Domain

@content
class SOA(Content):
    mname:     Domain
    rname:     Domain
    serialver: Int32
    refresh:   Int32
    retry:     Int32
    expire:    Int32
    minimum:   Int32

@content
class TXT(Content):
    text: SizedBytes[32]

@content
class A(Content):
    ip: Ipv4

@content
class AAAA(Content):
    ip: Ipv6

@content
class SRV(Content):
    priority: Int16
    weight:   Int16
    port:     Int16
    target:   Domain



"""
Sequence Field Implementation
"""
from dataclasses import InitVar, dataclass
from typing import ClassVar, Union, Type, get_origin
from typing_extensions import Self

from .. import Codec, Int32

#** Variables **#

#** Functions **#

def cname(seq: Union[object, type]) -> str:
    if isinstance(seq, type):
        return seq.__name__
    return seq.__class__.__name__

def is_datavar(anno: type) -> bool:
    origin = get_origin(anno)
    return origin in (ClassVar, InitVar)

def compile_annotation(name: str, anno: type, level: int = 1):
    """

    """
    # skip processing if already a codec or supported type
    if isinstance(anno, Codec):
        return anno
    # convert common types to defaults
    if anno == int:
        return Int32
    # convert property
    if isinstance(anno, Property):
        anno.hint = compile_annotation(name, anno.hint, level+1)
        return anno
    # raise error on unsupported type
    raise ValueError(f'invalid annotatation: {name!r}: {anno}')

#** Classes **#

@dataclass
class Property:
    hint: type

    def __class_getitem__(cls, hint: type) -> Self:
        return cls(hint)

@dataclass
class Field:
    name:     str
    type:     Type[Codec]
    dataattr: bool = True



"""
Standard UDP/TCP Client Implementations
"""
import random
import socket
from dataclasses import dataclass
from typing import List, Optional

from pypool import Pool
from pyserve import RawAddr

from . import BaseClient, Message

#** Variables **#
__all__ = ['UdpClient', 'TcpClient']

#** Classes **#

class SocketPool(Pool[socket.socket]):
    pass

@dataclass
class Client(BaseClient):
    addresses:  List[RawAddr]
    block_size: int           = 8192
    pool_size:  Optional[int] = None
    expiration: Optional[int] = None
    timeout:    int           = 30

    def __post_init__(self):
        self.pool = SocketPool(
            factory=self.newsock, 
            max_size=self.pool_size, 
            expiration=self.expiration)
    
    def newsock(self):
        raise NotImplementedError

    def pickaddr(self) -> RawAddr:
        """pick random address from list of addresses"""
        return random.choice(self.addresses)
    
    def drain(self):
        """drain socket pool"""
        self.pool.drain()

@dataclass
class UdpClient(Client):
   
    def newsock(self) -> socket.socket:
        """spawn new socket for the socket pool"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(self.timeout)
        return sock
    
#TODO: include some sort of UDP retry if response doesnt come back after timeout

    def request(self, msg: Message) -> Message:
        """
        send request to dns-server and recieve response
        """
        with self.pool.reserve() as sock:
            # send request
            addr = self.pickaddr()
            data = msg.encode()
            sock.sendto(data, addr)
            # recieve response
            data = sock.recv(self.block_size)
            return Message.decode(data)

class TcpClient(Client):

    def newsock(self) -> socket.socket:
        """spawn new socket for the socket pool"""
        addr = self.pickaddr()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect(addr)
        return sock

    def request(self, msg: Message) -> Message:
        """
        send request to dns-server and recieve response
        """
        with self.pool.reserve() as sock:
            # send request
            data = msg.encode()
            data = len(data).to_bytes(2, 'big') + data
            sock.send(data)
            # recieve size of response
            sizeb = sock.recv(2)
            size  = int.from_bytes(sizeb, 'big')
            # read data from size
            data = sock.recv(size)
            return Message.decode(data)




from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='pydns3',
    version='0.0.1',
    license='MIT',
    packages=find_packages(),
    url='https://github.com/imgurbot12/pydns',
    author='Andrew Scott',
    author_email='imgurbot12@gmail.com',
    description='Simple Python DNS Library. DNS Packet-Parsing/Client/Server',
    python_requires='>=3.7',
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=[
        'pypool',
        'pyserve',
        'dataclasses',
        'typing_extensions',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



"""
DNS Answer Record Instance/Format
"""
from typing import Type
from abc import abstractmethod
from dataclasses import dataclass
from typing import Protocol
from typing_extensions import Self

from .codec import *
from .content import Content
from .enum import RType, RClass
from . import content

#** Variables **#
__all__ = ['BaseAnswer', 'Answer', 'PreRequisite', 'Update']

#: all objects exported by content
CONTENT_ALL = [getattr(content, name) for name in content.__all__]

#: map of content-types mapped to content objects
CONTENT_MAP = {r.rtype:r for r in CONTENT_ALL if issubclass(r, Content)}

#** Functions **#

def get_rclass(rtype: RType) -> Type[Content]:
    """retrieve record content class based on record-type"""
    return CONTENT_MAP[rtype]

#** Classes **#

@make_sequence
class Header:
    name:   Domain
    rtype:  Int[16, RType, 'RType']
    rclass: Int[16, RClass, 'RClass']
    ttl:    Int32

class BaseAnswer(Protocol):
    """Baseclass for defining `Answer` objects"""
    name: bytes
    
    @property
    @abstractmethod
    def rtype(self) -> RType:
        raise NotImplementedError

    @abstractmethod
    def encode(self, ctx: Context) -> bytes:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def decode(cls, ctx: Context, raw: bytes) -> Self:
        raise NotImplementedError

@dataclass
class Answer(BaseAnswer):
    """Standard DNS Answer Implementation"""
    name:    bytes
    ttl:     int
    content: Content
    rclass:  RClass  = RClass.IN
    
    @property
    def rtype(self) -> RType:
        return self.content.rtype

    def encode(self, ctx: Context) -> bytes:
        head = Header(self.name, self.rtype, self.rclass, self.ttl).encode(ctx)
        ctx.index += 2 # pre-increment index since body-size goes before content
        body = self.content.encode(ctx)
        size = len(body).to_bytes(2, 'big')
        return head + size + body

    @classmethod
    def decode(cls, ctx: Context, raw: bytes) -> Self:
        # parse header and size of body
        header = Header.decode(ctx, raw)
        Int16.decode(ctx, raw)
        # determine content-type based on rtype in header
        rclass  = get_rclass(header.rtype)
        content = rclass.decode(ctx, raw)
        return cls(header.name, header.ttl, content, header.rclass)

@dataclass
class PreRequisite(Answer):
    """Alias for Answer in UPDATE action DNS Requests with Sensible Defaults"""
    ttl:     int = 0
    content: Content = field(default_factory=content.ANY)

class Update(Answer):
    """Alias of Answer in UPDATE action DNS Requests"""
    pass
